REBOL [
    Title:    "r80 v.5 -- lean webserver"
    Date:     8-Aug-2002
    Author:   "Viktor Pavlu"
    File:     %r80v5.r
    Home:     http://plain.at/vpavlu
    Email:    viktor_pavlu@hotmail.com
    Category: [ 'web 'tcp ]
    Version:  0.2.1
    History:  [ 0.0.1 {18-jan-2002 "rewriting of previous version of my REBserver"}
                0.0.2 {22-jan-2002 "added module support, all functionlity is in modules"}
                0.0.3 { 9-feb-2002 "severe debugging"}
                0.1.0 {22-mar-2002 "complete redesign, servlets must register for request patterns"}
                0.1.1 {23-mar-2002 "added server side scripting environment"}
                0.1.2 { 1-apr-2002 "removed some bugs"}
                0.2.0 { 2-aug-2002 "continued improving"}
                0.2.1 { 6-aug-2002 "data structures re-structured, apache compatible"}
    ]
    Comment: {
      this software is published under the NCSA open source license.
      see accompanying license.txt
      OSI Certified Open Source Software.
    }
]

;--- missing ---
; handling of POST data
; multithreaded execution

;-- config --
document-root: clean-path %./htdocs
access-log:    clean-path %access.log
module-dir:    clean-path %./modules/
;--

send-header: func [ r /local buf ][
  ;net-utils/export
  write-io r/conn buf: rejoin [
    "HTTP/1.0 " r/status crlf
    "Content-Type: " r/content-type crlf
    "Content-Length: " length? r/content crlf
    "Date: " to-idate now crlf 
    "^/"
  ] length? buf
]

dirname?: func [ "returns directory part of given path" path ][
  path: reduce-path path
  either #"/" = last to-string path [
    return to-file path
  ][
    return first split-path path 
  ]
]

suffix?: func [
    {Return the suffix (ext) of a filename or url, else NONE.}
    path [any-string!]
][
    if all [
        path: find/last path #"."
        not find path #"/"
    ] [to-file path]
]

sandbox: func [ "executes a handler, returns its output" 
                script [file!] "handler script (name only)"
                /local err ][
  if error? set/any 'err try [
    change-dir dirname? r/path-translated
    ;r/content: probe rejoin [ "do load " join module-dir script ]
    do load join module-dir script
  ][
    r/status:  200
    r/content: join "runtime error:^/" mold disarm err
  ]
  change-dir server/path
]

; noch nicht multithreaded...
process-query: func [ "gets appropriate handler for request and executes it" /local handler ][
  if not empty? processing-queue [
    r: first head processing-queue
    remove head processing-queue
    
    ;provide access to data through known objects
    system/options/cgi:  r/request
    ;system/ports/input:  r/conn
    ;system/ports/output: r/conn

    either exists? r/path-translated [
      handler: none
      foreach [ pattern file ] modules [
        if empty? to-string find/match/any r/request/script-name pattern [
          handler: copy file break
        ]
      ]
      
      either not none? handler [
        sandbox handler
      ][
        r/status: 200
        r/content: join "no handler found for: " r/request/script-name
      ]
    ][
      r/status: 200
      r/content: join "file not found: " r/request/script-name
    ]

    ;system/ports/input:  server/input
    ;system/ports/output: server/output

    send-header r
    either r/request/request-method = "HEAD" [ close r/conn ][ insert output-queue r ]
  ]
]

reduce-path: func [ "removes %. %.. and returns cleaned path" path /local t tokens ][
  path: trim to-string path
  if any [ empty? path "/" = path ][ return %/ ]
  tokens: parse/all path "/"
  while [ not tail? tokens ][
    t: first tokens
  
    either = t ""   [ remove tokens ][
    either = t "."  [ remove tokens ][
    either = t ".." [ tokens: back tokens remove/part tokens 2 ][
    tokens: next tokens
    ]]]
  ]
  tokens: head tokens
  if empty? tokens [ return %/ ]
  t: to-file copy first tokens
  tokens: next tokens
  forall tokens [ append t join "/" first tokens ]
  if (#"/" = last  path) [ append      t "/" ]
  if (#"/" = first path) [ insert head t "/" ]
  t
]

translate-path: func [ "maps a path to local document-root/" path ][
  if error? try [
    path: clean-path join document-root reduce-path path
    either none? find path document-root [ return document-root ][ return path ]
  ][
    return document-root
  ]
]

send-chunks: func [ "sends pending output to all clients" /local r tmp ][
  output-queue: head output-queue
  forall output-queue [
    r: first output-queue
    write-io r/conn tmp: copy/part r/content 1024 tmp: length? tmp
    byte-count: add byte-count tmp
    remove/part r/content tmp

    if empty? r/content [
      close r/conn
      remove output-queue
    ]
  ]
  output-queue: head output-queue
]

handle-new-connection: func [
  "checks incoming connection, prepares environment and appends to processing-queue"
  conn "incoming connection"
  /local buf blk r request-blk response-blk
][
  ;test connection
  if error? try [ buf: first conn ][ close conn return false ]
  blk: parse buf none
  
  if none? server/location [
    server/location: (rejoin [ http:// read join dns:// conn/local-ip ":" conn/local-port ])
  ]

  ;set up environment
  request-blk: make system/options/cgi compose [
    server-software:   (rejoin [ "r80v5/" system/script/header/version ])
    server-name:       (read dns://)
    gateway-interface: ("CGI/1.1")
    server-protocol:   ("HTTP/1.1")
    server-port:       (to-string conn/local-port)
    remote-addr:       (to-string conn/host)
  ]
  request-blk/request-method:  blk/1
  request-blk/server-protocol: blk/3

  ;GET data
  blk: parse/all blk/2 "?"
  request-blk/script-name: to-file reduce-path blk/1
  request-blk/query-string: blk/2

  ;POST data
  while [ ((buf: first conn) <> "") AND not none? buf ][
    set [ key value ] parse/all buf ":"
    repend request-blk/other-headers [ trim key trim value ]
  ]
  
  if not none? pos: find request-blk/other-headers "content-type" [
    request-blk/content-type: first next pos
    remove/part pos 2
    if not none? pos: find request-blk/other-headers "content-length" [
      request-blk/content-length: first next pos
      remove/part pos 2
    ]
  ]

  ;set up rest of environment
  r: make object! compose [
    conn:            (conn)
    request:         (request-blk)
    path-translated: (translate-path request-blk/script-name)
    status:          (200)
    content-type:    ("application/octet-stream")
    other-headers:   (copy [])
    content:         (copy "")
  ]
  append processing-queue r
]

;-- start --
secure none
system/options/quiet: yes
do load %conf.r

if all [ value? 'listen-port port? listen-port ][ close listen-port ]
listen-port: open/lines/direct join tcp://: port
;print join "listening on port " port

processing-queue: make block! 1024
output-queue: make block! 1024

if value? 'server [ unprotect 'server ]
server: make object! compose [
  location: none
  path:     (system/script/path)
;  sys-in:   (system/ports/input)
;  sys-out:  (system/ports/output)
]
protect 'server

byte-count: 0

forever [
  either all [
    empty? processing-queue
    empty? output-queue
  ][
    ;handle new connections
    wait listen-port
    either error? a: try [ conn: first listen-port ][
      close conn
    ][
      ;das is ja eigentlich dumm...
      either (length? processing-queue) > 100 [
        write-io conn buf: "HTTP/1.0 503 Server Overloaded^/" length? buf
        close conn
      ][
        handle-new-connection conn
      ]
    ]
  ][
    ;process pending queries
    if (length? output-queue) < 2 [ process-query ]
    send-chunks
  ]
  ;print join "bytes sent:" byte-count
]

